// %flair:license{
// This file is part of the Flair framework distributed under the
// CECILL-C License, Version 1.0.
// %flair:license}
//  created:    2013/08/23
//  filename:   NmeaGps.cpp
//
//  author:     Guillaume Sanahuja
//              Copyright Heudiasyc UMR UTC/CNRS 7253
//
//  version:    $Id: $
//
//  purpose:    Base class for GPS using NMEA sentances
//
//
/*********************************************************************/

#include "NmeaGps.h"
#include "geodesie.h"
#include <Euler.h>
#include <DataPlot1D.h>
#include <Tab.h>
#include <TabWidget.h>
#include <GridLayout.h>
#include <GroupBox.h>
#include <PushButton.h>
#include <FrameworkManager.h>
#include <Map.h>
#include <GeoCoordinate.h>
#include <Vector3D.h>
#include <Label.h>
#include <GpsData.h>
#include <string.h>

using std::string;
using namespace Geodesie;
using namespace flair::core;
using namespace flair::gui;

namespace flair {
namespace sensor {

NmeaGps::NmeaGps(const FrameworkManager *parent, string name, NMEAFlags_t NMEAFlags)
    : IODevice(parent, name) {
  this->NMEAFlags = NMEAFlags;

  nmea_zero_INFO(&info);
  nmea_parser_init(&parser);
  alt_ref = 0;

  if ((NMEAFlags & GGA) == 0) {
    Err("Enable at least the GGA sentence\n");
  }

 /*
  if ((NMEAFlags & GST) != 0) {
    desc->SetElementName(index, 0, "dev_lat");
    desc->SetElementName(index + 1, 0, "dev_lon");
    desc->SetElementName(index + 2, 0, "dev_elv");
    index += 3;
  }
  */

  // station sol
  main_tab = new Tab(parent->GetTabWidget(), name);
  tab = new TabWidget(main_tab->NewRow(), name);
  sensor_tab = new Tab(tab, "Reglages");
  GroupBox *reglages_groupbox = new GroupBox(sensor_tab->NewRow(), name);
  button_ref = new PushButton(reglages_groupbox->NewRow(), "set ref");
  nb_sat_label = new Label(reglages_groupbox->NewRow(), "nb_sat");
  fix_label = new Label(reglages_groupbox->LastRowLastCol(), "fix");

  position = new GeoCoordinate((IODevice *)this, "position", 0, 0, 0);

  take_ref = false;

  gpsData = new GpsData(this);
  AddDataToLog(gpsData);

  nb_sat_label->SetText("nb_sat: %i", gpsData->GetNumberOfSatellites());
  fix_label->SetText("fix: %i", gpsData->GetFixQuality());
}

NmeaGps::~NmeaGps() {
  nmea_parser_destroy(&parser);
  delete main_tab;
}

const GpsData *NmeaGps::GetDatas(void) const {
   return gpsData;
}

void NmeaGps::GetDatas(core::GpsData **outGpsData) const {
  *outGpsData = gpsData;
}

void NmeaGps::UseDefaultPlot(void) {
  plot_tab = new Tab(tab, "Mesures");

  if ((NMEAFlags & GGA) != 0) {
    e_plot = new DataPlot1D(plot_tab->NewRow(), "e", -10, 10);
    e_plot->AddCurve(gpsData->Element(GpsData::East));
    n_plot = new DataPlot1D(plot_tab->LastRowLastCol(), "n", -10, 10);
    n_plot->AddCurve(gpsData->Element(GpsData::North));
    u_plot = new DataPlot1D(plot_tab->LastRowLastCol(), "u", -10, 10);
    u_plot->AddCurve(gpsData->Element(GpsData::Up));
  }
  if ((NMEAFlags & VTG) != 0) {
    ve_plot = new DataPlot1D(plot_tab->NewRow(), "ve", -10, 10);
    ve_plot->AddCurve(gpsData->Element(GpsData::EastVelocity));
    vn_plot = new DataPlot1D(plot_tab->LastRowLastCol(), "vn", -10, 10);
    vn_plot->AddCurve(gpsData->Element(GpsData::NorthVelocity));
  }

  Tab *map_tab = new Tab(tab, "carte");
  map = new Map(map_tab->NewRow(), "map");
  map->AddPoint(position, "drone");
}

DataPlot1D *NmeaGps::EPlot(void) const {
  if ((NMEAFlags & GGA) != 0) {
    return e_plot;
  } else {
    Err("GGA sentence not requested\n");
    return NULL;
  }
}

DataPlot1D *NmeaGps::NPlot(void) const {
  if ((NMEAFlags & GGA) != 0) {
    return n_plot;
  } else {
    Err("GGA sentence not requested\n");
    return NULL;
  }
}

DataPlot1D *NmeaGps::UPlot(void) const {
  if ((NMEAFlags & GGA) != 0) {
    return u_plot;
  } else {
    Err("GGA sentence not requested\n");
    return NULL;
  }
}

DataPlot1D *NmeaGps::VEPlot(void) const {
  if ((NMEAFlags & VTG) != 0) {
    return ve_plot;
  } else {
    Err("GGA sentence not requested\n");
    return NULL;
  }
}

DataPlot1D *NmeaGps::VNPlot(void) const {
  if ((NMEAFlags & VTG) != 0) {
    return vn_plot;
  } else {
    Err("GGA sentence not requested\n");
    return NULL;
  }
}

Layout *NmeaGps::GetLayout(void) const { return sensor_tab; }

Tab *NmeaGps::GetPlotTab(void) const { return plot_tab; }

TabWidget *NmeaGps::GetTab(void) const { return tab; }

void NmeaGps::SetRef(void) { take_ref = true; }

void NmeaGps::GetEnu(Vector3D *point) {
  gpsData->GetMutex();
  gpsData->GetEnu(point->x,point->y,point->z);
  gpsData->ReleaseMutex();
}

void NmeaGps::parseFrame(const char *frame, int frame_size) {

  int result;

  result = nmea_parse(&parser, frame, frame_size, &info);
  if (result != 1) {
    Warn("unrecognized nmea sentence: %s\n",frame);
  }

  result = nmea_parse_GPGGA(frame, frame_size, &pack);

  if (result == 1) {
    nmea_info2pos(&info, &pos);
    // Printf("%s\n",frame);
   //  Printf("nb:%i fix:%i lat:%f long:%f alt:%f vel:%f angle:%f\n",pack.satinuse,pack.sig,info.lat,info.lon,info.elv,info.speed*1000./3600.,info.direction);
//Printf("lat:%f long:%f\n",pos.lat,pos.lon);

    gpsData->GetMutex(); // on utilise le mutex de gpsData pour fix et nb_sat
    if (gpsData->GetFixQuality() != (GpsData::FixQuality_t)pack.sig) {
      gpsData->SetFixQuality((GpsData::FixQuality_t)pack.sig);
      fix_label->SetText("fix: %i", pack.sig);
    }
    if (gpsData->GetNumberOfSatellites() != pack.satinuse) {
      gpsData->SetNumberOfSatellites(pack.satinuse) ;
      nb_sat_label->SetText("nb_sat: %i", pack.satinuse);
    }
    gpsData->ReleaseMutex();

    gpsData->SetLla(Euler::ToDegree(pos.lat), Euler::ToDegree(pos.lon),info.elv);
    position->SetCoordinates(Euler::ToDegree(pos.lat), Euler::ToDegree(pos.lon),
                             info.elv);

    if ((info.sig == 2 && alt_ref == 0) || button_ref->Clicked() == true ||
        take_ref == true) {
      Printf("prise pos ref\n");
      lat_ref = pos.lat;
      long_ref = pos.lon;
      alt_ref = info.elv;
      take_ref = false;
    }
    // if(alt_ref!=0)
    {
      double x, y, z;
      double e, n, u;
      Geographique_2_ECEF(pos.lon, pos.lat, info.elv, x, y, z);
      ECEF_2_ENU(x, y, z, e, n, u, long_ref, lat_ref, alt_ref);
      // Printf("lon:%f lat:%f elv:%f\n",pos.lon,pos.lat,info.elv);

      gpsData->SetEnu(e,n,u);

      if ((NMEAFlags & VTG) != 0) {
        gpsData->SetVelocity(info.speed * 1000. / 3600. * sin(Euler::ToRadian(info.direction)),
                                  info.speed * 1000. / 3600. * cos(Euler::ToRadian(info.direction)));
      }/*
      if ((NMEAFlags & GST) != 0) {
        // Thread::Printf("dev_lon:%f dev_lat:%f
        // dev_elv:%f\n",info.dev_lat,info.dev_lon,info.dev_elv);
        output->SetValueNoMutex(index, 0, info.dev_lat);
        output->SetValueNoMutex(index + 1, 0, info.dev_lon);
        output->SetValueNoMutex(index + 2, 0, info.dev_elv);
        index += 3;
      }*/


      gpsData->SetDataTime(GetTime());
      ProcessUpdate(gpsData);
    }
  }
}

} // end namespace sensor
} // end namespace framewor
